#include "roomlistmodel.h"

#include <algorithm>

#include <QDebug>

QModelIndex RoomListModel::calcIndex(std::string id, const std::vector<QSharedPointer<Room>> &invites,
                                     const std::vector<QSharedPointer<Room>> &other) {
    int pos = 0;
    for (const auto &r : invites) {
        if (r->id == id)
            return index(pos, 0);
        else
            pos++;
    }
    for (const auto &r : other) {
        if (r->id == id)
            return index(pos, 0);
        else
            pos++;
    }

    return index(-1, 0);
}

void RoomListModel::addRoom(QSharedPointer<Room> room) {
    int posIdx = (int)(invites.size() + joinedRooms.size());
    beginInsertRows(QModelIndex(), posIdx, posIdx);
    joinedRooms.push_back(room);
    endInsertRows();

    rooms[room->id] = room;
}

void RoomListModel::addInvite(QSharedPointer<Room> room) {
    beginInsertRows(QModelIndex(), (int)invites.size(), (int)invites.size());
    invites.push_back(room);
    endInsertRows();

    rooms[room->id] = room;
}

void RoomListModel::updateRoom(QSharedPointer<Room> room) {
    rooms[room->id] = room;

    auto idx = calcIndex(room->id, invites, joinedRooms);
    if (idx.row() != -1)
        emit dataChanged(idx, idx);
}

QHash<int, QByteArray> RoomListModel::roleNames() const {
    QHash<int, QByteArray> roles;
    roles[RoomModelRole] = "roomModel";
    roles[RoomNameRole] = "roomName";
    roles[UnreadNotificationsRole] = "unreadNotifications";
    roles[LastMessageRole] = "lastMessage";
    roles[RoomPictureRole] = "roomPicture";
    roles[TagRole] = "tag";
    roles[IsInvite] = "isInvite";
    return roles;
}

int RoomListModel::rowCount(const QModelIndex &parent) const {
    Q_UNUSED(parent);
    return (int)invites.size() + joinedRooms.size();
}

QVariant RoomListModel::data(const QModelIndex &index, int role) const {

    if (index.row() < 0 || index.row() >= rowCount())
        return QVariant();

    const auto &room = [this, index]() {
        if (index.row() < (int)invites.size())
            return invites[index.row()];
        else
            return joinedRooms[index.row() - invites.size()];
    }();

    switch (role) {
    case RoomModelRole:
        return QVariant::fromValue(room.data());
    case RoomNameRole:
        return QString::fromStdString(room->name());
    case UnreadNotificationsRole:
        return room->unreadNotifications;
    case LastMessageRole:
        return room->msgInfo.body;
    case RoomPictureRole:
        return QString::fromStdString(room->avatar_url).replace("mxc://", "image://MXCImage/");
    case TagRole:
        return QString::fromStdString(room->tags.empty() ? "" : room->tags.front());
    case IsInvite:
        return room->is_invite;

    default:
        return QVariant();
    }
}

RoomListModel::RoomListModel(QObject *parent) { Q_UNUSED(parent); }

bool RoomListSortProxy::lessThan(const QModelIndex &rLeft, const QModelIndex &rRight) const {
    QModelIndex const left_idx = sourceModel()->index(rLeft.row(), 0, QModelIndex());
    QModelIndex const right_idx = sourceModel()->index(rRight.row(), 0, QModelIndex());

    auto left = [this, left_idx](RoomListModel::RoomRoles role) -> QVariant {
        return sourceModel()->data(left_idx, role);
    };
    auto right = [this, right_idx](RoomListModel::RoomRoles role) -> QVariant {
        return sourceModel()->data(right_idx, role);
    };

    qDebug() << "left tag:" << left(RoomListModel::TagRole).toString()
             << " right tag: " << right(RoomListModel::TagRole).toString() << " compare result: "
             << QString::localeAwareCompare(left(RoomListModel::TagRole).toString(),
                                            right(RoomListModel::TagRole).toString());

    if (left(RoomListModel::IsInvite) != right(RoomListModel::IsInvite))
        return left(RoomListModel::IsInvite) > right(RoomListModel::IsInvite);
    else {
        if (left(RoomListModel::TagRole).toString() != right(RoomListModel::TagRole).toString()) {
            return QString::localeAwareCompare(left(RoomListModel::TagRole).toString(),
                                               right(RoomListModel::TagRole).toString()) < 0;
        } else {
            if (left(RoomListModel::UnreadNotificationsRole) != right(RoomListModel::UnreadNotificationsRole))
                return left(RoomListModel::UnreadNotificationsRole) > right(RoomListModel::UnreadNotificationsRole);
            else
                return rLeft.row() < rRight.row();
        }
    }
}
