#pragma once

#include <map>
#include <set>
#include <string>
#include <vector>

#include <QAbstractListModel>
#include <QColor>
#include <QDateTime>
#include <QString>

#include <mtx/events/collections.hpp>

struct DescInfo {
    QString event_id;
    QString username;
    QString userid;
    QString body;
    QString timestamp;
    QDateTime datetime;
};

struct MemberInfo {
    std::string avatar_url, display_name, userid;
};

class EventType : public QObject {
    Q_OBJECT

  public:
    EventType() : QObject() {}

    enum Type {
        // Unsupported event
        Unsupported,
        /// m.room_key_request
        KeyRequest,
        /// m.room.aliases
        Aliases,
        /// m.room.avatar
        Avatar,
        /// m.room.canonical_alias
        CanonicalAlias,
        /// m.room.create
        Create,
        /// m.room.encrypted.
        Encrypted,
        /// m.room.encryption.
        Encryption,
        /// m.room.guest_access
        GuestAccess,
        /// m.room.history_visibility
        HistoryVisibility,
        /// m.room.join_rules
        JoinRules,
        /// m.room.member
        Member,
        /// m.room.name
        Name,
        /// m.room.power_levels
        PowerLevels,
        /// m.room.topic
        Topic,
        /// m.room.redaction
        Redaction,
        /// m.room.pinned_events
        PinnedEvents,
        // m.sticker
        Sticker,
        // m.tag
        Tag,
        /// m.room.message
        AudioMessage,
        EmoteMessage,
        FileMessage,
        ImageMessage,
        LocationMessage,
        NoticeMessage,
        TextMessage,
        VideoMessage,
        UnknownMessage,
    };
    Q_ENUM(Type);
};

struct Room : public QAbstractListModel {
  private:
    Q_OBJECT
  public:
    explicit Room(QObject *parent = 0);

    enum Roles {
        Type,
        Body,
        FormattedBody,
        UserId,
        UserName,
        Timestamp,
        Url,
    };

    QHash<int, QByteArray> roleNames() const override;
    int rowCount(const QModelIndex &parent = QModelIndex()) const;
    QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const;

    Q_INVOKABLE QString userIdToUserName(QString id);
    Q_INVOKABLE QString roomName() const { return QString::fromStdString(name_); }
    Q_INVOKABLE QColor userColor(QString id, QColor background);

    std::string id, lastMessage, tag;

    int unreadNotifications = 0;

    //! The calculated name of the room.
    std::string name_;
    std::string canonical_alias;
    //! The topic of the room.
    std::string topic;
    //! The calculated avatar url of the room.
    std::string avatar_url;
    //! Whether or not the room is an invite.
    bool is_invite = false;
    //! Total number of members in the room.
    int16_t member_count = 0;
    //! Who can access to the room.
    mtx::events::state::JoinRule join_rule = mtx::events::state::JoinRule::Public;
    bool guest_access = false;
    //! Metadata describing the last message in the timeline.
    DescInfo msgInfo;
    //! The list of tags associated with this room
    std::vector<std::string> tags;

    std::set<std::string> members;
    std::set<std::string> eventIds;
    std::map<std::string, MemberInfo> memberInfos;

    void applyEvents(const std::vector<mtx::events::collections::TimelineEvents> &events, std::string prev_batch,
                     std::string next_batch) {
        emit newEvents(events, prev_batch, next_batch);
    }
    void applyTags(const std::vector<std::string> &tags) { emit setTags(tags); }
    void applyUnreadNotifications(int unread) { emit newUnreadState(unread); }
    std::string name();

  signals:
    void newEvents(std::vector<mtx::events::collections::TimelineEvents> events, std::string prev_batch,
                   std::string next_batch);
    void newTags(std::vector<std::string> tags);
    void newUnreadState(int);
    void roomNameChanged(QString name);

  public slots:
    void addEvents(std::vector<mtx::events::collections::TimelineEvents> events, std::string prev_batch,
                   std::string next_batch);
    void setTags(std::vector<std::string> tags) { this->tags = tags; }
    void setUnread(int unread) { this->unreadNotifications = unread; }
    void fetchOlderMessages();
    void sendTextMessage(QString message);
    void markRead(int index);

  private:
    std::vector<mtx::events::collections::TimelineEvents> events;
    std::string prev_batch;
    std::string next_batch;

    std::map<QString, QColor> userColors;
};
